@@ -11,12 +11,15 @@
 This module contains a function and a class for validating XSD content models,
 plus a set of functions for manipulating encoded content.
 """
-from collections import defaultdict, deque
-from typing import Any, Counter, Dict, Iterable, Iterator, List, \
+from collections import defaultdict, deque, Counter
+from operator import attrgetter
+from typing import Any, Dict, Iterable, Iterator, List, \
     MutableMapping, MutableSequence, Optional, Tuple, Union
+import warnings
 
 from ..exceptions import XMLSchemaValueError
-from ..aliases import ModelGroupType, ModelParticleType, SchemaElementType
+from ..aliases import ModelGroupType, ModelParticleType, SchemaElementType, \
+    OccursCounterType
 from ..translation import gettext as _
 from .. import limits
 from .exceptions import XMLSchemaModelError, XMLSchemaModelDepthError
@@ -27,6 +30,8 @@ AdvanceYieldedType = Tuple[ModelParticleType, int, List[SchemaElementType]]
 ContentItemType = Tuple[Union[int, str], Any]
 EncodedContentType = Union[MutableMapping[Union[int, str], Any], Iterable[ContentItemType]]
 
+get_occurs = attrgetter('min_occurs', 'max_occurs')
+
 
 def distinguishable_paths(path1: List[ModelParticleType], path2: List[ModelParticleType]) -> bool:
     """
@@ -184,13 +189,14 @@ class ModelVisitor:
     """
     _groups: List[Tuple[ModelGroupType, Iterator[ModelParticleType], bool]]
     element: Optional[SchemaElementType]
+    occurs: OccursCounterType
 
     __slots__ = '_groups', 'root', 'occurs', 'element', 'group', 'items', 'match'
 
     def __init__(self, root: ModelGroupType) -> None:
         self._groups = []
         self.root = root
-        self.occurs = Counter[Union[ModelParticleType, Tuple[ModelParticleType]]]()
+        self.occurs = Counter()
         self.element = None
         self.group = root
         self.items = self.iter_group()
@@ -226,26 +232,8 @@ class ModelVisitor:
 
     @property
     def expected(self) -> List[SchemaElementType]:
-        """
-        Returns the expected elements of the current and descendant groups.
-        """
-        expected: List[SchemaElementType] = []
-        items: Union[ModelGroupType, Iterator[ModelParticleType]]
-
-        if self.group.model == 'choice':
-            items = self.group
-        elif self.group.model == 'all':
-            items = (e for e in self.group if e.min_occurs > self.occurs[e])
-        else:
-            items = (e for e in self.group if e.min_occurs > self.occurs[e])
-
-        for e in items:
-            if isinstance(e, groups.XsdGroup):
-                expected.extend(e.iter_elements())
-            else:
-                expected.append(e)
-                expected.extend(e.maps.substitution_groups.get(e.name or '', ()))
-        return expected
+        """Returns the expected elements of the current and descendant groups."""
+        return self.group.get_expected(self.occurs)
 
     def restart(self) -> None:
         self.clear()
@@ -253,17 +241,34 @@ class ModelVisitor:
 
     def stop(self) -> Iterator[AdvanceYieldedType]:
         while self.element is not None:
-            for e in self.advance():
-                yield e
+            yield from self.advance()
 
     def iter_group(self) -> Iterator[ModelParticleType]:
         """Returns an iterator for the current model group."""
-        if self.group.max_occurs == 0:
-            return iter(())
-        elif self.group.model != 'all':
-            return iter(self.group)
+        if self.group.model == 'all':
+            for e in self.group.iter_elements():
+                if not e.is_over(self.occurs):
+                    yield e
+        elif self.group.max_occurs == 0:
+            return
         else:
-            return (e for e in self.group.iter_elements() if not e.is_over(self.occurs[e]))
+            yield from self.group.content
+
+    def match_element(self, tag: str) -> Optional[SchemaElementType]:
+        if self.element is None:
+            raise XMLSchemaValueError("cannot match, %r is ended!" % self)
+        elif self.element.max_occurs == 0:
+            return None
+        elif self.element.name is None:
+            return self.element.match(tag, group=self.root, occurs=self.occurs)
+        elif tag == self.element.name:
+            return self.element
+        else:
+            for xsd_element in self.element.iter_substitutes():
+                if tag == xsd_element.name:
+                    return xsd_element
+            else:
+                return None
 
     def advance(self, match: bool = False) -> Iterator[AdvanceYieldedType]:
         """
@@ -272,168 +277,300 @@ class ModelVisitor:
 
         :param match: provides current element match.
         """
-        def stop_item(item: ModelParticleType) -> bool:
+        item: ModelParticleType
+        item_occurs: int
+
+        def stop_item() -> bool:
             """
             Stops element or group matching, incrementing current group counter.
 
             :return: `True` if the item has violated the minimum occurrences for itself \
             or for the current group, `False` otherwise.
             """
+            nonlocal item
+            nonlocal item_occurs
+
+            item_occurs = occurs[item]
             if isinstance(item, groups.XsdGroup):
                 self.group, self.items, self.match = self._groups.pop()
 
             if self.group.model == 'choice':
-                item_occurs = occurs[item]
                 if not item_occurs:
                     return False
-                item_max_occurs = occurs[(item,)] or item_occurs
 
-                if item.max_occurs is None:
-                    min_group_occurs = 1
-                elif item_occurs % item.max_occurs:
-                    min_group_occurs = 1 + item_occurs // item.max_occurs
+                high_occurs = occurs[item.oid] or item_occurs
+                min_occurs, max_occurs = get_occurs(item)
+
+                if max_occurs is None:
+                    occurs[self.group] += 1
+                elif item_occurs % max_occurs:
+                    occurs[self.group] += 1 + item_occurs // max_occurs
                 else:
-                    min_group_occurs = item_occurs // item.max_occurs
+                    occurs[self.group] += item_occurs // max_occurs
 
-                max_group_occurs = max(1, item_max_occurs // (item.min_occurs or 1))
-
-                occurs[self.group] += min_group_occurs
-                occurs[(self.group,)] += max_group_occurs
-                occurs[item] = 0
+                occurs[self.group.oid] += (high_occurs // (min_occurs or 1)) or 1
 
+                occurs[item] = occurs[item.oid] = 0
                 self.items = self.iter_group()
                 self.match = False
-                return item.is_missing(item_max_occurs)
+                return min_occurs > high_occurs  # type: ignore[no-any-return]
 
             elif self.group.model == 'all':
-                return False
+                return False  # 'all' models can only be checked at the end
             elif self.match:
                 pass
-            elif occurs[item]:
+            elif item_occurs:
                 self.match = True
             elif item.is_emptiable():
                 return False
             elif self._groups:
-                return stop_item(self.group)
-            elif self.group.min_occurs <= max(occurs[self.group], occurs[(self.group,)]):
-                return stop_item(self.group)
-            else:
+                item = self.group
+                return stop_item()
+            elif self.group.is_missing(occurs):
                 return True
+            else:
+                item = self.group
+                return stop_item()
 
-            if item is self.group[-1]:
-                for k, item2 in enumerate(self.group, start=1):  # pragma: no cover
-                    item_occurs = occurs[item2]
-                    if not item_occurs:
+            if item is self.group.content[-1]:
+                for k, item2 in enumerate(self.group.content, start=1):  # pragma: no cover
+                    low_occurs = occurs[item2]
+                    if not low_occurs:
                         continue
 
-                    item_max_occurs = occurs[(item2,)] or item_occurs
-                    if item_max_occurs == 1 or any(not x.is_emptiable() for x in self.group[k:]):
+                    high_occurs = occurs[item2.oid] or low_occurs
+                    if high_occurs == 1 or \
+                            any(not x.is_emptiable() for x in self.group.content[k:]):
                         occurs[self.group] += 1
+                        occurs[self.group.oid] += 1
                         break
 
-                    min_group_occurs = max(1, item_occurs // (item2.max_occurs or item_occurs))
-                    max_group_occurs = max(1, item_max_occurs // (item2.min_occurs or 1))
-
-                    occurs[self.group] += min_group_occurs
-                    occurs[(self.group,)] += max_group_occurs
+                    occurs[self.group] += (low_occurs // (item2.max_occurs or low_occurs)) or 1
+                    occurs[self.group.oid] += (high_occurs // (item2.min_occurs or 1)) or 1
                     break
 
-            return item.is_missing(max(occurs[item], occurs[(item,)]))
+            return item.is_missing(occurs)
+
+        def model_error_tuple() -> AdvanceYieldedType:
+            if occurs[item]:
+                expected = item.get_expected(occurs)
+            else:
+                occurs[item] = item_occurs
+                expected = item.get_expected(occurs)
+                occurs[item] = 0
+
+            return item, item_occurs, expected
 
-        occurs = self.occurs
         if self.element is None:
             raise XMLSchemaValueError("cannot advance, %r is ended!" % self)
 
+        item = self.element
+        occurs = self.occurs
+        item_occurs = occurs[item]
+
         if match:
-            occurs[self.element] += 1
+            occurs[item] += 1
             self.match = True
             if self.group.model == 'all':
-                self.items = (e for e in self.group.iter_elements() if not e.is_over(occurs[e]))
-            elif not self.element.is_over(occurs[self.element]):
-                return
-            elif self.group.model == 'choice' and self.element.is_ambiguous():
+                self.items = self.iter_group()
+            elif not item.is_over(occurs) or \
+                    self.group.model == 'choice' and item.is_ambiguous():
                 return
 
-        obj = None
         try:
-            element_occurs = occurs[self.element]
-            if stop_item(self.element):
-                yield self.element, element_occurs, [self.element]
+            if stop_item():
+                yield model_error_tuple()
 
             while True:
-                while self.group.is_over(max(occurs[self.group], occurs[(self.group,)])):
-                    stop_item(self.group)
+                while self.group.is_over(occurs):
+                    item = self.group
+                    stop_item()
 
-                obj = next(self.items, None)
-                if isinstance(obj, groups.XsdGroup):
-                    # inner 'sequence' or 'choice' XsdGroup
-                    self._groups.append((self.group, self.items, self.match))
-                    self.group = obj
-                    self.items = self.iter_group()
-                    self.match = False
-                    occurs[obj] = occurs[(obj,)] = 0
-
-                elif obj is not None:
-                    # XsdElement or XsdAnyElement
-                    self.element = obj
-                    if self.group.model == 'sequence':
-                        occurs[obj] = 0
-                    return
-
-                elif not self.match:
-                    if self.group.model == 'all':
-                        if all(e.min_occurs <= occurs[e] for e in self.group.iter_elements()):
-                            occurs[self.group] = 1
-
-                    group, expected = self.group, self.expected
-                    if stop_item(group) and expected:
-                        yield group, occurs[group], expected
-
-                elif self.group.model != 'all':
-                    self.items, self.match = self.iter_group(), False
-                elif any(e.min_occurs > occurs[e] for e in self.group.iter_elements()):
-                    if not self.group.min_occurs:
-                        yield self.group, occurs[self.group], self.expected
-                    self.group, self.items, self.match = self._groups.pop()
-                elif any(not e.is_over(occurs[e]) for e in self.group):
-                    self.items = self.iter_group()
-                    self.match = False
+                for obj in self.items:
+                    if isinstance(obj, groups.XsdGroup):
+                        # inner 'sequence' or 'choice' XsdGroup
+                        self._groups.append((self.group, self.items, self.match))
+                        self.group = obj
+                        self.items = self.iter_group()
+                        self.match = False
+                        occurs[obj] = occurs[obj.oid] = 0
+                        break
+                    else:
+                        # XsdElement or XsdAnyElement
+                        self.element = obj
+                        if self.group.model == 'sequence':
+                            occurs[obj] = 0
+                        return
                 else:
-                    occurs[self.group] = 1
+                    if self.match:
+                        self.items, self.match = self.iter_group(), False
+                    elif self.group.model == 'all':
+                        self.group, self.items, self.match = self._groups.pop()
+                    else:
+                        item = self.group
+                        if stop_item():
+                            yield model_error_tuple()
 
         except IndexError:
             # Model visit ended
             self.element = None
-            if self.group.is_missing(max(occurs[self.group], occurs[(self.group,)])):
-                if self.group.model == 'choice':
-                    yield self.group, occurs[self.group], self.expected
-                elif self.group.model == 'sequence':
-                    if obj is not None:
-                        yield self.group, occurs[self.group], self.expected
-                elif any(e.min_occurs > occurs[e] for e in self.group):
-                    yield self.group, occurs[self.group], self.expected
-            elif self.group.max_occurs is not None and self.group.max_occurs < occurs[self.group]:
+            if self.group.model == 'all':
+                yield from self._iter_all_model_errors(occurs)
+            elif self.group.is_missing(occurs) or self.group.is_exceeded(occurs):
                 yield self.group, occurs[self.group], self.expected
 
+    def _iter_all_model_errors(self, occurs: OccursCounterType) -> Iterator[AdvanceYieldedType]:
+        """Validate occurrences in an 'all' model, yielding error tuples."""
+        stack: List[Tuple[groups.XsdGroup, Iterator[ModelParticleType]]] = []
+        group, particles = self.group, iter(self.group)
+        zero_missing: List[Tuple[groups.XsdGroup, ModelParticleType]] = []
+
+        while True:
+            for item in particles:
+                if occurs[item]:
+                    occurs[group] = 1
+
+                if isinstance(item, groups.XsdGroup):
+                    if item.max_occurs == 0:
+                        continue
+
+                    stack.append((group, particles))
+                    particles = iter(item.content)
+                    if len(stack) > limits.MAX_MODEL_DEPTH:
+                        raise XMLSchemaModelDepthError(self.group)
+                    break
+
+                if item.is_missing(occurs) or item.is_exceeded(occurs):
+                    if occurs[item]:
+                        yield item, occurs[item], item.get_expected(occurs)
+                    else:
+                        zero_missing.append((group, item))
+            else:
+                if group.is_missing(occurs) or group.is_exceeded(occurs):
+                    if occurs[group] or not stack:
+                        yield group, occurs[group], group.get_expected(occurs)
+                    else:
+                        zero_missing.append((stack[-1][0], group))
+
+                if not stack:
+                    break
+                group, particles = stack.pop()
+
+        # Late check on missing items that never occurs
+        for group, item in zero_missing:
+            if occurs[group]:
+                yield item, occurs[item], item.get_expected(occurs)
+
     # Kept for backward compatibility
     def iter_unordered_content(
             self, content: EncodedContentType,
             default_namespace: Optional[str] = None) -> Iterator[ContentItemType]:
-        return iter_unordered_content(content, self.root, default_namespace)
+
+        msg = f"{self.__class__.__name__}.iter_unordered_content() method will " \
+              "be removed in v4.0, use iter_unordered_content() function instead."
+        if default_namespace is not None:
+            msg += " Don't provide default_namespace argument, it's ignored."
+        warnings.warn(msg, DeprecationWarning, stacklevel=2)
+
+        return iter_unordered_content(content, self.root)
 
     def iter_collapsed_content(
             self, content: Iterable[ContentItemType],
             default_namespace: Optional[str] = None) -> Iterator[ContentItemType]:
-        return iter_collapsed_content(content, self.root, default_namespace)
+
+        msg = f"{self.__class__.__name__}.iter_collapsed_content() method will " \
+              "be removed in v4.0, use iter_collapsed_content() function instead."
+        if default_namespace is not None:
+            msg += " Don't provide default_namespace argument, it's ignored."
+        warnings.warn(msg, DeprecationWarning, stacklevel=2)
+
+        return iter_collapsed_content(content, self.root)
+
+
+class InterleavedModelVisitor(ModelVisitor):
+    """
+    A visitor for openContent interleaved models. Memorizes an internal state
+    for deciding when to advance the model. The model doesn't advance if the
+    last match_element() call is with the wildcard.
+    """
+    __slots__ = 'wildcard', '_advance_model'
+
+    def __init__(self, root: ModelGroupType, wildcard: XsdAnyElement) -> None:
+        super().__init__(root)
+        self.wildcard = wildcard
+        self._advance_model = True
+        if self.element is None:
+            self.element = wildcard
+
+    def clear(self) -> None:
+        super().clear()
+        self._advance_model = True
+        if self.element is None:
+            self.element = self.wildcard
+
+    def match_element(self, tag: str) -> Optional[SchemaElementType]:
+        xsd_element = super().match_element(tag)
+        if xsd_element is not None or self.element is self.wildcard:
+            return xsd_element
+        elif not self.wildcard.is_matching(tag, group=self.root, occurs=self.occurs):
+            return None
+
+        for xsd_element in self.group.iter_elements():
+            if xsd_element.is_matching(tag, group=self.root, occurs=self.occurs):
+                if not xsd_element.is_over(self.occurs):
+                    return None
+        else:
+            if self.wildcard.process_contents != 'strict' or tag in self.root.maps.elements:
+                self._advance_model = False
+                return self.wildcard
+            return None
+
+    def advance(self, match: bool = False) -> Iterator[AdvanceYieldedType]:
+        if self.element is None:
+            yield from super().advance(match)
+        elif self.element is self.wildcard:
+            if not match:
+                self.element = None
+        elif not self._advance_model:
+            self._advance_model = True
+        else:
+            yield from super().advance(match)
+            if self.element is None:
+                self.element = self.wildcard
+
+
+class SuffixedModelVisitor(ModelVisitor):
+    """A visitor for openContent suffixed models."""
+
+    __slots__ = 'wildcard',
+
+    def __init__(self, root: ModelGroupType, wildcard: XsdAnyElement) -> None:
+        super().__init__(root)
+        self.wildcard = wildcard
+        if self.element is None:
+            self.element = wildcard
+
+    def clear(self) -> None:
+        super().clear()
+        if self.element is None:
+            self.element = self.wildcard
+
+    def advance(self, match: bool = False) -> Iterator[AdvanceYieldedType]:
+        if self.element is None:
+            yield from super().advance(match)
+        elif self.element is not self.wildcard:
+            yield from super().advance(match)
+            if self.element is None:
+                self.element = self.wildcard
+        elif not match:
+            self.element = None
 
 
 #
 # Functions for manipulating encoded content
 
-def iter_unordered_content(
-        content: EncodedContentType,
-        group: ModelGroupType,
-        default_namespace: Optional[str] = None) -> Iterator[ContentItemType]:
+def iter_unordered_content(content: EncodedContentType, group: ModelGroupType) \
+        -> Iterator[ContentItemType]:
     """
     Takes an unordered content stored in a dictionary of lists and yields the
     content elements sorted with the ordering defined by the model group. Character
@@ -449,7 +586,6 @@ def iter_unordered_content(
     dictionary the values must be lists where each item is the content \
     of a single element.
     :param group: the model group related to content.
-    :param default_namespace: the default namespace to apply for matching names.
     """
     consumable_content: Dict[str, Any]
 
@@ -499,16 +635,13 @@ def iter_unordered_content(
         yield cdata_content.pop()
 
 
-def sort_content(content: EncodedContentType,
-                 group: ModelGroupType,
-                 default_namespace: Optional[str] = None) -> List[ContentItemType]:
-    return [x for x in iter_unordered_content(content, group, default_namespace)]
+def sort_content(content: EncodedContentType, group: ModelGroupType) \
+        -> List[ContentItemType]:
+    return [x for x in iter_unordered_content(content, group)]
 
 
-def iter_collapsed_content(
-        content: Iterable[ContentItemType],
-        group: ModelGroupType,
-        default_namespace: Optional[str] = None) -> Iterator[ContentItemType]:
+def iter_collapsed_content(content: Iterable[ContentItemType], group: ModelGroupType) \
+        -> Iterator[ContentItemType]:
     """
     Iterates a content stored in a sequence of couples *(name, value)*, yielding
     items in the same order of the sequence, except for repetitions of the same
@@ -518,11 +651,10 @@ def iter_collapsed_content(
 
     This iteration mode, in cooperation with the method *iter_encode* of the class
     XsdGroup, facilitates the encoding of content formatted with a convention that
-    collapses the children with the same tag into a list (eg. BadgerFish).
+    collapses the children with the same tag into a list (e.g. BadgerFish).
 
     :param content: an iterable containing couples of names and values.
     :param group: the model group related to content.
-    :param default_namespace: the default namespace to apply for matching names.
     """
     prev_name = None
     unordered_content: Dict[str, Any] = defaultdict(deque)
@@ -564,7 +696,7 @@ def iter_collapsed_content(
             yield name, value
             prev_name = name
 
-    # Add the remaining consumable content onto the end of the data.
+    # Yields the remaining consumable content after the end of the data.
     for name, values in unordered_content.items():
         for v in values:
             yield name, v
